-module(kv_vnode_manager).

-behaviour(gen_server).

-export([
    start_link/0,
    get_local/2,
    put_local/3
]).

%%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2]).

-record(kv_vnode_manager, {table :: ets:table()}).

-type state() :: #kv_vnode_manager{}.

%%% public api

start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

-spec get_local(Index :: non_neg_integer(), Key :: term()) -> {ok, Value :: term()} | none.
get_local(Index, Key) ->
    gen_server:call(?MODULE, {get, Index, Key}).

-spec put_local(Index :: non_neg_integer(), Key :: term(), Value :: term()) -> ok.
put_local(Index, Key, Value) ->
    gen_server:cast(?MODULE, {put, Index, Key, Value}).

%%% gen_server callbacks

-spec init([]) -> {ok, state()}.
init([]) ->
    {ok, #kv_vnode_manager{table = ets:new(?MODULE, [set])}}.

handle_call({get, Index, Key}, _From, State) ->
    VNode = get_vnode(Index, State),
    {reply, kv_vnode:get(VNode, Key), State}.

handle_cast({put, Index, Key, Value}, State) ->
    VNode = get_vnode(Index, State),
    ok = kv_vnode:put(VNode, Key, Value),
    {noreply, State}.

%%% internal functions

get_vnode(Index, #kv_vnode_manager{table = Table}) ->
    case ets:lookup(Table, Index) of
        [{_, Pid}] ->
            Pid;
        [] ->
            {ok, Pid} = kv_vnode_sup:start_vnode(Index),
            ets:insert(Table, {Index, Pid}),
            Pid
    end.
